/* ------------------------------------------------------------------------
 * Copyright 2013 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */
package org.adjective.syntactic.convert.j8to7;

import org.adjective.syntactic.convert.j8to7.type.MethodInfo;
import org.adjective.syntactic.convert.j8to7.type.TypeInfo;
import org.adjective.syntactic.convert.j8to7.util.ClassFinder;
import org.adjective.syntactic.convert.j8to7.util.Importer;
import org.adjective.syntactic.parser.ast.ASTAllocationExpression;
import org.adjective.syntactic.parser.ast.ASTArguments;
import org.adjective.syntactic.parser.ast.ASTClassOrInterfaceBody;
import org.adjective.syntactic.parser.ast.ASTClassOrInterfaceBodyElement;
import org.adjective.syntactic.parser.ast.ASTFormalParameters;
import org.adjective.syntactic.parser.ast.ASTMember;
import org.adjective.syntactic.parser.ast.ASTMethodDeclaration;
import org.adjective.syntactic.parser.ast.ASTModifiers;
import org.adjective.syntactic.parser.ast.ASTNameList;
import org.adjective.syntactic.parser.name.ParameterizedName;
import org.adjective.syntactic.parser.name.SimpleParameterizedName;
import org.adjective.syntactic.parser.name.SimpleTypeParameter;
import org.adjective.syntactic.parser.name.TypeParameter;
import org.adjective.syntactic.parser.node.ExpressionNode;
import org.adjective.syntactic.parser.type.JavaType;
import org.adjective.syntactic.parser.type.SimpleJavaType;
import org.adjective.syntactic.parser.util.ModifierSet;
import org.adjective.syntactic.util.ArrayUtil;
import org.adjective.syntactic.util.JavaTypeUtil;

import java.util.Arrays;
import java.util.Collection;

public class AbstractASTConverter
{
    private final Importer _importer;
    private final ClassFinder _classFinder;

    public AbstractASTConverter(final Importer importer, final ClassFinder classFinder)
    {
        _importer = importer;
        _classFinder = classFinder;
    }

    protected TypeInfo getTypeInfo(final JavaType forType)
    {
        final String name = _importer.getQualifiedName(forType);
        final TypeInfo typeInfo = _classFinder.getClass(name, getTypeParameters(forType));
        if (typeInfo == null)
        {
            throw new ConversionException("Cannot find '" + name + "' in " + _classFinder);
        }
        return typeInfo;
    }

    private TypeParameter[] getTypeParameters(final JavaType forType)
    {
        return ArrayUtil.last(forType.getParameterizedTypeName()).getParameters();
    }

    protected ExpressionNode makeAnonymousInnerClass(final JavaType interfaceType, final MethodInfo functionalMethod,
                                                     final ASTMethodDeclaration methodImpl)
    {
        final ModifierSet set = new ModifierSet(ModifierSet.Modifier.PUBLIC);
        final ASTModifiers modifiers = new ASTModifiers(set);
        final ASTMember member = new ASTMember(modifiers, methodImpl);
        final JavaType astType = substituteType(methodImpl.getParameters(), functionalMethod, interfaceType);
        ASTClassOrInterfaceBody body = new ASTClassOrInterfaceBody(new ASTClassOrInterfaceBodyElement(member));
        return new ASTAllocationExpression(JavaTypeUtil.withoutWildcard(astType), new ASTArguments(), body);
    }

    private JavaType substituteType(final ASTFormalParameters parameters, final MethodInfo method, final JavaType type)
    {
        final Integer[] typeParameters = method.getClassTypeParameters();
        if (ArrayUtil.isNull(typeParameters))
        {
            return type;
        }
        final ParameterizedName lastName = ArrayUtil.last(type.getParameterizedTypeName());
        TypeParameter[] classParameters = Arrays.copyOf(lastName.getParameters(), lastName.getParameters().length);
        for (int i = 0; i < typeParameters.length; i++)
        {
            final Integer classIndex = typeParameters[i];
            if (classIndex != null)
            {
                classParameters[classIndex] = new SimpleTypeParameter(TypeParameter.Kind.EXACT,
                                                                      parameters.getParameterTypes()[i]);
            }
        }

        ParameterizedName[] name = new ParameterizedName[type.getParameterizedTypeName().length];
        System.arraycopy(type.getParameterizedTypeName(), 0, name, 0, name.length - 1);
        name[name.length - 1] = new SimpleParameterizedName(lastName.getName(), classParameters);
        return new SimpleJavaType(name, type.getArrayDepth());
    }

    protected MethodInfo getFunctionalMethod(final JavaType forType)
    {
        final TypeInfo typeInfo = getTypeInfo(forType);
        return getFunctionalMethod(typeInfo);
    }

    private MethodInfo getFunctionalMethod(final TypeInfo typeInfo)
    {
        final Collection<MethodInfo> methods = typeInfo.getAbstractMethods();
        if (methods.size() != 1)
        {
            throw new ConversionException("Class '" + typeInfo.getName()
                                                  + "' should have exactly 1 method but has " + methods.size());
        }

        return methods.iterator().next();
    }

    protected TypeInfo findType(final String name)
    {
        String qualifiedName = _importer.getQualifiedName(name);
        if (qualifiedName == null)
        {
            qualifiedName = name;
        }
        final TypeInfo typeInfo = _classFinder.getClass(qualifiedName, new TypeParameter[0]);
        return typeInfo;
    }

    protected ASTNameList getExceptions(final MethodInfo method)
    {
        return method.getExceptionTypes().length == 0 ? null
                : new ASTNameList(method.getExceptionTypes());
    }
}
