/*
 * ------------------------------------------------------------------------
 *  Copyright 2013 Tim Vernum
 *  ------------------------------------------------------------------------
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  ------------------------------------------------------------------------
 */
package org.adjective.syntactic.convert.j8to7;

import org.adjective.syntactic.convert.j8to7.util.ClassFinder;
import org.adjective.syntactic.convert.j8to7.util.ClassPath;
import org.adjective.syntactic.convert.j8to7.util.Importer;
import org.adjective.syntactic.convert.j8to7.util.VariableStack;
import org.adjective.syntactic.parser.ast.*;
import org.adjective.syntactic.parser.node.BaseNode;
import org.adjective.syntactic.parser.node.ExpressionNode;
import org.adjective.syntactic.parser.node.ExpressionSuffix;
import org.adjective.syntactic.parser.node.Nodes;
import org.adjective.syntactic.parser.node.StatementNode;
import org.adjective.syntactic.parser.node.VariableDeclarationNode;
import org.adjective.syntactic.parser.type.JavaType;
import org.adjective.syntactic.parser.util.DefaultVisitor;
import org.adjective.syntactic.parser.util.ModifierSet;
import org.adjective.syntactic.util.Pair;

import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

public class Convert8To7Visitor extends DefaultVisitor<Object, Object>
{
    private final Stack<String> _className;
    private final VariableStack _vars;
    private final ClassFinder _classFinder;
    private final Importer _importer;
    private final LambdaConverter _lambdaConverter;
    private final MemberReferenceConverter _referenceConverter;
    private final List<Pair<VariableDeclarationNode, String>> _toMarkFinal;

    public Convert8To7Visitor()
    {
        _className = new Stack<String>();
        _vars = new VariableStack();
        _classFinder = new ClassFinder();
        _importer = new Importer(_classFinder);
        _lambdaConverter = new LambdaConverter(_importer, _classFinder);
        _referenceConverter = new MemberReferenceConverter(_importer, _classFinder);
        _toMarkFinal = new ArrayList<Pair<VariableDeclarationNode, String>>();
    }

    public void setClassPath(ClassPath cp)
    {
        _classFinder.setClassLoader(new URLClassLoader(cp.getUrls()));
    }

    public void convert(final ASTCompilationUnit tree)
    {
        if (tree == null)
        {
            throw new IllegalArgumentException("No AST supplied");
        }
        tree.jjtAccept(this, null);
    }

    @Override
    public Object visit(final ASTCompilationUnit node, final Object data)
    {
        final Object val = super.visit(node, data);
        for (Pair<VariableDeclarationNode, String> pair : _toMarkFinal)
        {
            markFinal(pair.tail, pair.head);
        }
        return val;
    }

    @Override
    public Object visit(final ASTImportDeclaration node, final Object data)
    {
        _importer.add(node);
        return super.visit(node, data);
    }

    private void markFinal(final String name, final VariableDeclarationNode node)
    {
        if (node.isMultiVariable())
        {
            Map<String, VariableDeclarationNode> split = node.split();
            ASTBlock block = ((BaseNode) node).findAncestor(ASTBlock.class);
            split.get(name).getModifiers().set(ModifierSet.Modifier.FINAL);
            List<StatementNode> statements = new ArrayList<StatementNode>();
            for (VariableDeclarationNode decl : split.values())
            {
                assert decl instanceof ASTLocalVariableDeclaration;
                statements.add(new ASTVariableDeclarationStatement((ASTLocalVariableDeclaration) decl));
            }
            block.replace(node.jjtGetParent(), statements);
        }
        else
        {
            node.getModifiers().set(ModifierSet.Modifier.FINAL);
        }
    }

    @Override
    public Object visit(final ASTClassOrInterfaceDeclaration node, final Object data)
    {
        _className.push(node.getName().getIdentifier());
        _classFinder.define(node);
        try
        {
            return super.visit(node, data);
        }
        finally
        {
            _className.pop();
        }
    }

    @Override
    public Object visit(final ASTMethodDeclaration node, final Object data)
    {
        _vars.pushScope();
        try
        {
            return super.visit(node, data);
        }
        finally
        {
            _vars.popScope();
        }
    }

    @Override
    public Object visit(final ASTBlock node, final Object data)
    {
        _vars.pushScope();
        try
        {
            return super.visit(node, data);
        }
        finally
        {
            _vars.popScope();
        }
    }

    @Override
    public Object visit(final ASTFormalParameter node, final Object data)
    {
        _vars.store(node.getName().getIdentifier(), node);
        return super.visit(node, data);
    }

    @Override
    public Object visit(final ASTLocalVariableDeclaration node, final Object data)
    {
        for (ASTVariableDeclarator var : node.getVariables())
        {
            _vars.store(var.getIdentifier().getIdentifier(), node);

            if (!var.hasInitializer())
            {
                continue;
            }

            final ExpressionNode init = var.getInitializer();
            if (!init.is(ASTExpression.class))
            {
                continue;
            }

            ExpressionNode expr = init.as(ASTExpression.class).getBaseExpression();
            if (!expr.is(ASTPrimaryExpression.class))
            {
                continue;
            }

            final ASTPrimaryExpression primary = expr.as(ASTPrimaryExpression.class);
            ExpressionNode prefix = primary.getPrefix();
            if (prefix.is(ASTLambdaExpression.class))
            {
                final ASTLambdaExpression lambda = prefix.as(ASTLambdaExpression.class);
                final ExpressionNode convert = _lambdaConverter.convertLambda(lambda, getType(node, var));
                fixBody(lambda.getBody());
                var.replaceInitializer(new ASTExpression(convert));
                continue;
            }

            final Nodes<ExpressionSuffix> suffixes = primary.getSuffixes();
            if (suffixes.has(ASTMemberReference.class))
            {
                final List<ExpressionSuffix> leading = new ArrayList<ExpressionSuffix>();
                for (ExpressionSuffix suffix : suffixes)
                {
                    if (suffix.is(ASTMemberReference.class))
                    {
                        final ASTMemberReference reference = suffix.as(ASTMemberReference.class);
                        final ExpressionNode target = makePrimaryExpression(prefix, leading);
                        final ExpressionNode convert = _referenceConverter.convertReference(target,
                                                                                            reference,
                                                                                            getType(node, var));
                        prefix = new ASTBracketExpression(convert);
                        leading.clear();
                    }
                    else
                    {
                        leading.add(suffix);
                    }
                }
                var.replaceInitializer(makePrimaryExpression(prefix, leading));
            }
        }

        return super.visit(node, data);
    }

    private ExpressionNode makePrimaryExpression(final ExpressionNode prefix, final List<ExpressionSuffix> suffix)
    {
        return suffix.isEmpty() ? prefix : new ASTPrimaryExpression(prefix, suffix);
    }

    private void fixBody(final ASTLambdaBody body)
    {
        body.jjtAccept(new DefaultVisitor<Object, Object>()
        {
            @Override
            public Object visit(final ASTPrimaryExpression node, final Object data)
            {
                final ExpressionNode prefix = node.getPrefix();
                if (node.jjtGetNumChildren() == 1)
                {
                    if (prefix.is(ASTQualifiedIdentifier.class))
                    {
                        String id = prefix.as(ASTQualifiedIdentifier.class).getName();
                        markFinal(id);
                    }
                }
                if (prefix.is(ASTThis.class))
                {
                    final ASTQualifiedIdentifier newPrefix = new ASTQualifiedIdentifier(_className.peek());
                    final ASTMemberSuffix newSuffix = new ASTMemberSuffix(prefix.as(ASTThis.class));
                    node.setTo(newPrefix, merge(newSuffix, node.getSuffixes()));
                }
                return super.visit(node, data);
            }

        }, null);
    }

    private void markFinal(final String var)
    {
        final VariableDeclarationNode node = _vars.get(var);
        if (node == null)
        {
            return;
        }
        _toMarkFinal.add(new Pair<VariableDeclarationNode, String>(node, var));
    }

    private <T> Iterable<T> merge(final T first, final Iterable<T> rest)
    {
        final ArrayList<T> list = new ArrayList<T>();
        list.add(first);
        for (T t : rest)
        {
            list.add(t);
        }
        return list;
    }

    private JavaType getType(ASTLocalVariableDeclaration declaration, ASTVariableDeclarator variable)
    {
        return BaseNode.makeType(declaration.getTypeNode(), variable.getArraySuffix());
    }

}
