/* ------------------------------------------------------------------------
 * Copyright 2013 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */
package org.adjective.syntactic.convert.j8to7;

import org.adjective.syntactic.convert.j8to7.type.MethodInfo;
import org.adjective.syntactic.convert.j8to7.util.ClassFinder;
import org.adjective.syntactic.convert.j8to7.util.Importer;
import org.adjective.syntactic.parser.ast.*;
import org.adjective.syntactic.parser.node.ExpressionNode;
import org.adjective.syntactic.parser.type.JavaType;

import java.util.Iterator;

public class LambdaConverter extends AbstractASTConverter
{
    public LambdaConverter(final Importer importer, final ClassFinder classFinder)
    {
        super(importer, classFinder);
    }

    public ExpressionNode convertLambda(final ASTLambdaExpression lambda, final JavaType forType)
    {
        final MethodInfo methodInfo = getFunctionalMethod(forType);
        final ASTMethodDeclaration astMethod = makeMethodDeclaration(methodInfo,
                                                                     lambda.getParameters(),
                                                                     lambda.getBody());
        return makeAnonymousInnerClass(forType, methodInfo, astMethod);
    }

    private ASTMethodDeclaration makeMethodDeclaration(final MethodInfo method, final ASTLambdaParameters parameters, final ASTLambdaBody lambdaBody)
    {
        final ASTMethodBody body = getMethodBody(lambdaBody);
        final ASTNameList exceptions = getExceptions(method);
        return new ASTMethodDeclaration(method.getReturnType(),
                                        new ASTIdentifier(method.getName()),
                                        makeFormalParameters(method, parameters),
                                        new ASTArraySuffixList(),
                                        exceptions,
                                        body);
    }

    private ASTFormalParameters makeFormalParameters(final MethodInfo method, final ASTLambdaParameters parameters)
    {
        final JavaType[] parameterTypes;
        if (parameters.hasDeclaredParameterTypes())
        {
            parameterTypes = parameters.getParameterListNode().getParameterTypes();
        }
        else
        {
            parameterTypes = method.getParameterTypes();
        }
        final String[] parameterNames = new String[parameterTypes.length];
        final Iterator<ASTIdentifier> iterator = parameters.getParameterNames().iterator();
        for (int i = 0; i < parameterNames.length; i++)
        {
            if (iterator.hasNext())
            {
                parameterNames[i] = iterator.next().getIdentifier();
            }
            else
            {
                throw new ConversionException("Incorrect number of arguments to " + method + " - " + parameterNames.length + " expected");
            }
        }
        ASTFormalParameter[] formalParameters = new ASTFormalParameter[parameterTypes.length];
        for (int i = 0; i < formalParameters.length; i++)
        {
            formalParameters[i] = new ASTFormalParameter(parameterTypes[i],
                                                         i == formalParameters.length - 1 && method.isVarArgs(),
                                                         parameterNames[i]);
        }
        return new ASTFormalParameters(formalParameters);
    }

    private ASTMethodBody getMethodBody(final ASTLambdaBody lambdaBody)
    {
        ASTBlock block;
        if (lambdaBody.isExpression())
        {
            ASTReturnStatement stmt = new ASTReturnStatement(lambdaBody.getExpression());
            block = new ASTBlock(stmt);
        }
        else
        {
            block = lambdaBody.getBlock();
        }
        return new ASTMethodBody(block);
    }
}
