/* ------------------------------------------------------------------------
 * Copyright 2013 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */

package org.adjective.syntactic.convert.j8to7;

import org.adjective.syntactic.convert.j8to7.util.ClassPath;
import org.adjective.syntactic.parser.jj.JavaParser;
import org.adjective.syntactic.parser.node.BaseNode;
import org.adjective.syntactic.parser.util.PrintSourceVisitor;
import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;

public class Main
{
    @Option(name = "--source", aliases = {"-s"}, usage = "Source directory or file", required = true)
    private File _source;

    @Option(name = "--output", aliases = {"-d", "-o"}, usage = "Output directory or file", required = true)
    private File _output;

    @Option(name = "--classpath", aliases = {"-cp"}, usage = "Classpath for references classes", required = false)
    private String _classpath;

    @SuppressWarnings("FieldCanBeLocal")
    @Option(name = "--flatten", aliases = {"-f"}, usage = "Flatten directory structure", required = false)
    private boolean _flatten = false;

    public static void main(String[] args) throws CmdLineException, IOException
    {
        final Main main = new Main();
        final CmdLineParser parser = new CmdLineParser(main);
        parser.setUsageWidth(120);
        parser.parseArgument(args);
        main.execute();
    }

    public void execute() throws IOException
    {
        if (!_source.exists())
        {
            throw new FileNotFoundException("No such file or directory: " + _source.getAbsolutePath());
        }
        if (_source.isFile())
        {
            processFile(_source, _output);
        }
        else if (_source.isDirectory())
        {
            processDirectory(_source, _output);
        }
    }

    private void processDirectory(final File source, final File output) throws IOException
    {
        final File[] files = source.listFiles();
        if (files == null)
        {
            return;
        }
        for (File file : files)
        {
            File out = new File(output, file.getName());
            if (file.isDirectory())
            {
                if (_flatten)
                {
                    out = output;
                }
                if (!out.isDirectory() && !out.mkdir())
                {
                    throw new FileNotFoundException("Cannot create directory " + out);
                }
                processDirectory(file, out);
            }
            else if (file.getName().endsWith(".java"))
            {
                processFile(file, out);
            }
        }
    }

    private void processFile(final File source, File output) throws IOException
    {
        if (output.isDirectory())
        {
            output = new File(output, source.getName());
        }
        BaseNode ast = parse(source);
        if (ast == null)
        {
            return;
        }
        PrintStream out = new PrintStream(new FileOutputStream(output));
        try
        {
            final Convert8To7Visitor converter = new Convert8To7Visitor();
            if (_classpath != null)
            {
                converter.setClassPath(new ClassPath(_classpath));
            }
            ast.jjtAccept(converter, null);
            ast.jjtAccept(new PrintSourceVisitor(), out);
        }
        finally
        {
            out.close();
        }
    }

    private static BaseNode parse(final File source)
    {
        try
        {
            return JavaParser.process(source);
        }
        catch (RuntimeException e)
        {
            System.err.println("Cannot parse: " + source.getPath());
            throw e;
        }
    }
}
