/* ------------------------------------------------------------------------
 * Copyright 2013 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */
package org.adjective.syntactic.convert.j8to7;

import org.adjective.syntactic.convert.j8to7.type.MethodInfo;
import org.adjective.syntactic.convert.j8to7.type.TypeInfo;
import org.adjective.syntactic.convert.j8to7.util.ClassFinder;
import org.adjective.syntactic.convert.j8to7.util.Importer;
import org.adjective.syntactic.parser.ast.*;
import org.adjective.syntactic.parser.node.ExpressionNode;
import org.adjective.syntactic.parser.node.ExpressionSuffix;
import org.adjective.syntactic.parser.node.StatementNode;
import org.adjective.syntactic.parser.type.BaseType;
import org.adjective.syntactic.parser.type.JavaType;
import org.adjective.syntactic.parser.type.SimpleJavaType;
import org.adjective.syntactic.parser.util.ModifierSet;

import java.util.Arrays;

public class MemberReferenceConverter extends AbstractASTConverter
{

    public MemberReferenceConverter(final Importer importer, final ClassFinder classFinder)
    {
        super(importer, classFinder);
    }

    public ExpressionNode convertReference(final ExpressionNode target, final ASTMemberReference reference, final JavaType forType)
    {
        final MethodInfo methodInfo = getFunctionalMethod(forType);
        final ASTMethodDeclaration astMethod = makeMethodDeclaration(methodInfo, target, reference);
        return makeAnonymousInnerClass(forType, methodInfo, astMethod);
    }

    private ASTMethodDeclaration makeMethodDeclaration(final MethodInfo method, final ExpressionNode target, final ASTMemberReference reference)
    {
        final ASTFormalParameters parameters = makeFormalParameters(method);
        final ASTNameList exceptions = getExceptions(method);

        ExpressionNode expression = null;
        if (target instanceof ASTQualifiedIdentifier)
        {
            // Might be a class reference [static member ref, receiver member ref, or constructor ref]
            final TypeInfo refType = findType(target.as(ASTQualifiedIdentifier.class).getName());
            if (refType != null)
            {
                expression = makeMethodBody(method, refType, reference);
            }
        }
        if (expression == null)
        {
            expression = makeMethodBody(method, target, reference);
        }

        final StatementNode statement;
        if (method.getReturnType().getBaseType() == BaseType.VOID)
        {
            statement = new ASTExpressionStatement(expression);
        }
        else
        {
            statement = new ASTReturnStatement(expression);
        }
        final ASTMethodBody body = new ASTMethodBody(new ASTBlock(statement));
        return new ASTMethodDeclaration(method.getReturnType(),
                                        new ASTIdentifier(method.getName()),
                                        parameters,
                                        new ASTArraySuffixList(),
                                        exceptions,
                                        body);
    }

    private ExpressionNode makeMethodBody(final MethodInfo method, final ExpressionNode target, final ASTMemberReference reference)
    {
        final ASTArguments arguments = getArguments(method);
        final String methodName = reference.getName().getIdentifier();
        return makeMethodCall(target, methodName, arguments);
    }

    private ExpressionNode makeMethodBody(final MethodInfo method, final TypeInfo refType, final ASTMemberReference reference)
    {
        final SimpleJavaType javaType = new SimpleJavaType(refType.getName());
        final ASTArguments arguments = getArguments(method);
        if (reference.isConstructorReference())
        {
            return new ASTAllocationExpression(javaType, arguments, null);
        }
        else
        {
            final String methodName = reference.getName().getIdentifier();
            MethodInfo refMethod = refType.findMethod(methodName, arguments.getNumberOfArguments());
            if (refMethod != null && refMethod.getModifiers().is(ModifierSet.Modifier.STATIC))
            {
                return makeStaticMethodCall(javaType, refMethod, arguments);
            }
            if (arguments.getNumberOfArguments() > 0)
            {
                refMethod = refType.findMethod(methodName, arguments.getNumberOfArguments() - 1);
                if (refMethod != null)
                {
                    return makeMethodCall(arguments.firstArgument(), refMethod, arguments.shift());
                }
            }
            throw new ConversionException("Cannot find method " + methodName + " in " + refType);
        }
    }

    private ExpressionNode makeStaticMethodCall(final JavaType javaType, final MethodInfo refMethod, final ASTArguments arguments)
    {
        final ExpressionNode prefix = new ASTQualifiedIdentifier(javaType.getTypeName());
        return makeMethodCall(prefix, refMethod, arguments);
    }

    private ExpressionNode makeMethodCall(final ExpressionNode prefix, final MethodInfo refMethod, final ASTArguments arguments)
    {
        final String methodName = refMethod.getName();
        return makeMethodCall(prefix, methodName, arguments);
    }

    private ExpressionNode makeMethodCall(final ExpressionNode prefix, final String methodName, final ASTArguments arguments)
    {
        final ASTIdentifierSuffix member = new ASTIdentifierSuffix(new ASTIdentifier(methodName));
        return new ASTPrimaryExpression(prefix,
                                        Arrays.<ExpressionSuffix>asList(new ASTMemberSuffix(member), arguments));
    }

    private ASTArguments getArguments(final MethodInfo method)
    {
        final String[] parameterNames = method.getParameterNames();
        final ASTQualifiedIdentifier[] argNames = new ASTQualifiedIdentifier[parameterNames.length];
        for (int i = 0; i < parameterNames.length; i++)
        {
            argNames[i] = new ASTQualifiedIdentifier(parameterNames[i]);
        }
        return new ASTArguments(new ASTArgumentList(argNames));
    }

    private ASTFormalParameters makeFormalParameters(final MethodInfo method)
    {
        final JavaType[] parameterTypes = method.getParameterTypes();
        final String[] parameterNames = method.getParameterNames();
        final ASTFormalParameter[] formalParameters = new ASTFormalParameter[parameterTypes.length];
        for (int i = 0; i < formalParameters.length; i++)
        {
            formalParameters[i] = new ASTFormalParameter(parameterTypes[i],
                                                         i == formalParameters.length - 1 && method.isVarArgs(),
                                                         parameterNames[i]);
        }
        return new ASTFormalParameters(formalParameters);
    }
}
