/* ------------------------------------------------------------------------
 * Copyright 2013 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */
package org.adjective.syntactic.convert.j8to7.type;

import org.adjective.syntactic.parser.ast.ASTMethodDeclaration;
import org.adjective.syntactic.parser.ast.ASTNameList;
import org.adjective.syntactic.parser.type.JavaType;
import org.adjective.syntactic.parser.util.ModifierSet;

public class ASTMethodInfo extends AbstractMethodInfo implements MethodInfo
{
    private final ASTMethodDeclaration _method;
    private final TypeParameterLookup _parameterLookup;
    private final ModifierSet _modifiers;

    public ASTMethodInfo(final ModifierSet modifiers, final ASTMethodDeclaration method, final TypeParameterLookup lookup)
    {
        if (modifiers == null)
        {
            throw new IllegalArgumentException("[Internal Error] Modifiers cannot be null");
        }
        if (method == null)
        {
            throw new IllegalArgumentException("[Internal Error] Method cannot be null");
        }
        if (lookup == null)
        {
            throw new IllegalArgumentException("[Internal Error] Lookup cannot be null");
        }
        _method = method;
        _parameterLookup = lookup;
        _modifiers = new ModifierSet(modifiers);
        if (isDeclaredOnObject())
        {
            _modifiers.clear(ModifierSet.Modifier.ABSTRACT);
        }
    }

    private JavaType getType(final JavaType type)
    {
        final JavaType typeParameter = _parameterLookup.findParameter(type.getTypeName());
        if (typeParameter != null)
        {
            return typeParameter;
        }
        else
        {
            return type;
        }
    }

    private JavaType[] getTypes(final JavaType[] in)
    {
        JavaType[] out = new JavaType[in.length];
        for (int i = 0; i < in.length; i++)
        {
            out[i] = getType(in[i]);
        }
        return out;
    }

    @Override
    public JavaType getReturnType()
    {
        return getType(_method.getActualReturnType());
    }

    @Override
    public String getName()
    {
        return _method.getName().getIdentifier();
    }

    @Override
    public ModifierSet getModifiers()
    {
        return _modifiers;
    }

    @Override
    public JavaType[] getParameterTypes()
    {
        return getTypes(_method.getParameters().getParameterTypes());
    }

    @Override
    public String[] getParameterNames()
    {
        return _method.getParameters().getParameterNames();
    }

    @Override
    public Integer[] getClassTypeParameters()
    {
        final JavaType[] types = _method.getParameters().getParameterTypes();
        final Integer[] vals = new Integer[types.length];

        for (int i = 0; i < vals.length; i++)
        {
            final JavaType type = types[i];
            if (type == null)
            {
                throw new IllegalStateException("[Internal Error] AST Method " + _method.getDebugInfo()
                                                        + " has no type for parameter " + i);
            }
            final Integer lookup = _parameterLookup.findParameterIndex(type.getTypeName());
            if (lookup != null)
            {
                vals[i] = lookup;
            }
        }

        return vals;
    }

    @Override
    public boolean isVarArgs()
    {
        return _method.getParameters().isVarArgs();
    }

    @Override
    public JavaType[] getExceptionTypes()
    {
        final ASTNameList exceptions = _method.getThrows();
        return exceptions == null ? new JavaType[0] : exceptions.getTypes();
    }

}
