/*
 * ------------------------------------------------------------------------
 *  Copyright 2013 Tim Vernum
 *  ------------------------------------------------------------------------
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  ------------------------------------------------------------------------
 */
package org.adjective.syntactic.convert.j8to7.type;

import org.adjective.syntactic.parser.type.JavaType;
import org.adjective.syntactic.parser.util.ModifierSet;

import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;

public class ClassMethodInfo extends AbstractMethodInfo implements MethodInfo
{
    private final Method _method;
    private final TypeParameterLookup _parameterLookup;

    public ClassMethodInfo(final Method method, final TypeParameterLookup parameterLookup)
    {
        _method = method;
        _parameterLookup = parameterLookup;
    }

    public static ClassMethodInfo[] array(final Method[] methods, TypeParameterLookup parameterLookup)
    {
        final ClassMethodInfo[] info = new ClassMethodInfo[methods.length];
        for (int i = 0; i < methods.length; i++)
        {
            info[i] = new ClassMethodInfo(methods[i], parameterLookup);
        }
        return info;
    }

    @Override
    public JavaType getReturnType()
    {
        return getJavaType(_method.getGenericReturnType(), _method.getReturnType());
    }

    @Override
    public String getName()
    {
        return _method.getName();
    }

    @Override
    public ModifierSet getModifiers()
    {
        final ModifierSet modifiers = new ModifierSet(_method.getModifiers());
        if (isDeclaredOnObject())
        {
            modifiers.clear(ModifierSet.Modifier.ABSTRACT);
        }
        return modifiers;
    }

    @Override
    public JavaType[] getParameterTypes()
    {
        return getJavaTypes(_method.getGenericParameterTypes(), _method.getParameterTypes());
    }

    private JavaType[] getJavaTypes(final Type[] genericTypes, final Class<?>[] classes)
    {
        final JavaType[] javaTypes = new JavaType[genericTypes.length];
        for (int i = 0; i < genericTypes.length; i++)
        {
            javaTypes[i] = getJavaType(genericTypes[i], classes[i]);
        }
        return javaTypes;
    }

    private JavaType getJavaType(final Type parameter, final Class<?> cls)
    {
        final JavaType javaType;
        if (parameter instanceof TypeVariable)
        {
            javaType = lookup((TypeVariable) parameter);
        }
        else
        {
            javaType = new ClassType(cls);
        }
        return javaType;
    }

    private JavaType lookup(final TypeVariable var)
    {
        final JavaType type = this._parameterLookup.findParameter(var.getName());
        return type != null ? type : new ClassType(Object.class);
    }

    @Override
    public String[] getParameterNames()
    {
        final String[] names = new String[_method.getParameterTypes().length];
        for (int i = 0; i < names.length; i++)
        {
            names[i] = "arg" + i;
        }
        return names;
    }

    @Override
    public Integer[] getClassTypeParameters()
    {
        final Type[] genericTypes = _method.getGenericParameterTypes();
        final Integer[] index = new Integer[genericTypes.length];
        for (int i = 0; i < index.length; i++)
        {
            Type parameter = genericTypes[i];
            if (parameter instanceof TypeVariable)
            {
                final TypeVariable var = (TypeVariable) parameter;
                index[i] = _parameterLookup.findParameterIndex(var.getName());
            }
        }
        return index;
    }

    @Override
    public boolean isVarArgs()
    {
        return _method.isVarArgs();
    }

    @Override
    public JavaType[] getExceptionTypes()
    {
        return getJavaTypes(_method.getGenericExceptionTypes(), _method.getExceptionTypes());
    }

    @Override
    public String toString()
    {
        return _method.toString();
    }
}
