/*
 * ------------------------------------------------------------------------
 *  Copyright 2013 Tim Vernum
 *  ------------------------------------------------------------------------
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  ------------------------------------------------------------------------
 */
package org.adjective.syntactic.convert.j8to7.util;

import org.adjective.syntactic.convert.j8to7.type.ASTTypeInfo;
import org.adjective.syntactic.convert.j8to7.type.ClassInfo;
import org.adjective.syntactic.convert.j8to7.type.TypeInfo;
import org.adjective.syntactic.parser.ast.ASTClassOrInterfaceDeclaration;
import org.adjective.syntactic.parser.name.TypeParameter;

import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Map;

public class ClassFinder
{
    private final Map<String, ASTClassOrInterfaceDeclaration> _localTypes;
    private ClassLoader _classLoader;

    public ClassFinder()
    {
        _localTypes = new HashMap<String, ASTClassOrInterfaceDeclaration>();
        _classLoader = getClass().getClassLoader();
    }

    public void setClassLoader(ClassLoader classLoader)
    {
        this._classLoader = classLoader;
    }

    public boolean hasClass(final String qualifiedName)
    {
        return findClass(qualifiedName) != null;
    }

    public TypeInfo getClass(final String qualifiedName, TypeParameter[] parameters)
    {
        final Class<?> cls = findClass(qualifiedName);
        if (cls == null)
        {
            ASTClassOrInterfaceDeclaration local = _localTypes.get(qualifiedName);
            if (local == null)
            {
                return null;
            }
            return new ASTTypeInfo(local, parameters);
        }
        return new ClassInfo(cls, parameters);
    }

    private Class<?> findClass(String qualifiedName)
    {
        for (; ; )
        {
            final Class<?> cls = loadClass(qualifiedName);
            if (cls != null)
            {
                return cls;
            }
            int dot = qualifiedName.lastIndexOf('.');
            if (dot == -1)
            {
                return null;
            }
            qualifiedName = qualifiedName.substring(0, dot) + '$' + qualifiedName.substring(dot + 1);
        }
    }

    private Class<?> loadClass(final String qualifiedName)
    {
        try
        {
            return _classLoader.loadClass(qualifiedName);
        }
        catch (ClassNotFoundException e)
        {
            return null;
        }
    }

    public void define(final ASTClassOrInterfaceDeclaration node)
    {
        _localTypes.put(node.getName().getIdentifier(), node);
    }

    @Override
    public String toString()
    {
        return getClass().getSimpleName() + "//" + describe(_classLoader);
    }

    private CharSequence describe(final ClassLoader classLoader)
    {
        if (classLoader instanceof URLClassLoader)
        {
            URLClassLoader ucl = (URLClassLoader) classLoader;
            StringBuilder description = new StringBuilder(ucl.getClass().getSimpleName());
            description.append('~');
            for (URL url : ucl.getURLs())
            {
                description.append(url);
                description.append(',');
            }
            return description;
        }
        return classLoader.toString();
    }
}
