/*
 * ------------------------------------------------------------------------
 *  Copyright 2013 Tim Vernum
 *  ------------------------------------------------------------------------
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  ------------------------------------------------------------------------
 */

package org.adjective.syntactic.parser.ast;

import org.adjective.syntactic.parser.node.BaseNode;
import org.adjective.syntactic.parser.node.ExpressionNode;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class ASTFloatingPointLiteral extends BaseNode implements ExpressionNode
{
    public ASTFloatingPointLiteral(int id)
    {
        super(id);
    }

    public <R, T> R jjtAccept(JavaParserVisitor<R, T> visitor, T data)
    {
        return visitor.visit(this, data);
    }

    public static class FloatingPointValue
    {
        public final BigDecimal base;
        public final BigInteger exponent;
        public final boolean isDouble;

        public FloatingPointValue(final BigDecimal base, final BigInteger exponent, final boolean isDouble)
        {
            this.base = base;
            this.exponent = exponent;
            this.isDouble = isDouble;
        }

        @Override
        public String toString()
        {
            StringBuilder builder = new StringBuilder(this.base.toString());
            if (this.exponent != null)
            {
                builder.append("e").append(exponent);
            }
            if (isDouble)
            {
                builder.append("d");
            }
            else
            {
                builder.append("f");
            }
            return builder.toString();
        }
    }

    public FloatingPointValue getValue()
    {
        String image = jjtGetFirstToken().image.toLowerCase();
        boolean dbl = true;
        if (image.endsWith("d"))
        {
            image = image.substring(0, image.length() - 1);
        }
        else if (image.endsWith("f"))
        {
            dbl = false;
            image = image.substring(0, image.length() - 1);
        }
        if (image.startsWith("0x"))
        {
            return parseHex(image.substring(2), dbl);
        }
        else
        {
            return parseDecimal(image, dbl);
        }
    }

    private static final Pattern DEC = Pattern.compile("([0-9]*\\.?[0-9]*)(?:e([-+]?[0-9]+))?");

    static FloatingPointValue parseDecimal(final String image, final boolean dbl)
    {
        final Matcher matcher = DEC.matcher(image);
        if (!matcher.matches())
        {
            throw new IllegalArgumentException("Not a valid floating point literal: " + image);
        }
        String expString = matcher.group(2);
        if (expString != null && expString.startsWith("+"))
        {
            expString = expString.substring(1);
        }
        BigInteger exp = expString == null ? null : new BigInteger(expString);
        BigDecimal value = new BigDecimal(matcher.group(1));
        return new FloatingPointValue(value, exp, dbl);
    }

    private static final Pattern HEX = Pattern.compile("([0-9a-f]*\\.[0-9a-f]*)p([-+]?[0-9]+)");

    static FloatingPointValue parseHex(final String image, final boolean dbl)
    {
        final Matcher matcher = HEX.matcher(image);
        if (!matcher.matches())
        {
            throw new IllegalArgumentException("Not a valid floating point literal: " + image);
        }
        String expString = matcher.group(2);
        if (expString.startsWith("+"))
        {
            expString = expString.substring(1);
        }
        BigInteger exp = new BigInteger(expString);
        BigDecimal value;
        final String val = "0x" + matcher.group(1) + "p0";
        if (dbl)
        {
            value = new BigDecimal(Double.valueOf(val));
        }
        else
        {
            value = new BigDecimal(Float.valueOf(val));
        }
        return new FloatingPointValue(value, exp, dbl);
    }
}
