/* ------------------------------------------------------------------------
 * Copyright 2013 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */
package org.adjective.syntactic.parser.type;

import org.adjective.syntactic.parser.ast.ASTIdentifier;
import org.adjective.syntactic.parser.ast.ASTQualifiedIdentifier;
import org.adjective.syntactic.parser.name.ParameterizedName;
import org.adjective.syntactic.parser.name.UnparameterizedName;

public class SimpleJavaType implements JavaType
{
    private final String _name;
    private final ParameterizedName[] _parameterizedName;
    private final int _arrayDepth;

    public SimpleJavaType(final ASTQualifiedIdentifier name)
    {
        this(name.getName(), parameterize(name), 0);
    }
    public SimpleJavaType(final String name)
    {
        this(name, parameterize(name), 0);
    }

    public SimpleJavaType(final ParameterizedName[] name, int arrayDepth)
    {
        this(combine(name), name, arrayDepth);
    }

    private SimpleJavaType(final String name, ParameterizedName[] parameterizedName, int arrayDepth)
    {
        _name = name;
        _parameterizedName = parameterizedName;
        _arrayDepth = arrayDepth;
    }

    private static ParameterizedName[] parameterize(final ASTQualifiedIdentifier name)
    {
        final ParameterizedName[] parameterizedName = new ParameterizedName[name.jjtGetNumChildren()];
        int i = 0;
        for (ASTIdentifier identifier : name.getIdentifierComponents())
        {
            parameterizedName[i++] = new UnparameterizedName(identifier);
        }
        return parameterizedName;
    }

    private static ParameterizedName[] parameterize(final String name)
    {
        final String[] split = name.split("\\.");
        final ParameterizedName[] parameterizedName = new ParameterizedName[split.length];
        for (int i = 0; i < parameterizedName.length; i++)
        {
            parameterizedName[i] = new UnparameterizedName(split[i]);
        }
        return parameterizedName;
    }

    private static String combine(final ParameterizedName[] name)
    {
        StringBuilder builder = new StringBuilder(name[0].getName());
        for (int i = 1; i < name.length; i++)
        {
            builder.append(".");
            builder.append(name[i].getName());
        }
        return builder.toString();
    }

    @Override
    public BaseType getBaseType()
    {
        return BaseType.OBJECT;
    }

    @Override
    public String getTypeName()
    {
        return _name;
    }

    @Override
    public ParameterizedName[] getParameterizedTypeName()
    {
        return _parameterizedName;
    }

    @Override
    public int getArrayDepth()
    {
        return _arrayDepth;
    }
}
