/* ------------------------------------------------------------------------
 * Copyright 2012 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */

package org.adjective.syntactic.parser.util;

import org.adjective.syntactic.parser.node.BaseNode;
import org.adjective.syntactic.parser.ast.Node;
import org.adjective.syntactic.parser.ast.SimpleNode;

import java.io.PrintStream;

/**
 * @author <a href="http://blog.adjective.org/">Tim Vernum</a>
 */
public class ASTPrinter
{
    private final PrintStream _stream;
    private boolean _showSource;
    private boolean _showTree;

    public ASTPrinter()
    {
        this(System.out);
    }

    public ASTPrinter(PrintStream stream)
    {
        _stream = stream;
    }

    public boolean showSource()
    {
        return _showSource || !_showTree;
    }

    public boolean showTree()
    {
        return _showTree || !_showSource;
    }

    public void print(String file, Node node)
    {
        String prefix = "";

        if (showTree())
        {
            print((SimpleNode) node, prefix);
        }

        if (showSource())
        {
            try
            {
                node.jjtAccept(new PrintSourceVisitor(), System.out);
            }
            catch (UnsupportedOperationException e)
            {
                System.err.println("\n(( In " + file + ": Cannot convert node to source: " + e.getMessage() + "))");
                System.err.println("\t" + e.getStackTrace()[0]);
            }
            catch(RuntimeException e) {
                System.err.println("\nException in " + file);
                e.printStackTrace();
            }
        }
    }

    private void print(SimpleNode node, String prefix)
    {
        _stream.print(node.toString(prefix));
        if (node.jjtGetNumChildren() == 0 || getChild(node, 0).jjtGetFirstToken() != node.jjtGetFirstToken())
        {
            _stream.print(' ');
            _stream.print(node.jjtGetFirstToken().image);
        }
        _stream.print('\n');
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
        {
            SimpleNode child = getChild(node, i);
            print(child, prefix + "  ");
        }
    }

    private BaseNode getChild(SimpleNode node, int i)
    {
        return (BaseNode) node.jjtGetChild(i);
    }

    public void setShowSource(final boolean showSource)
    {
        this._showSource = showSource;
    }

    public void setShowTree(final boolean showTree)
    {
        this._showTree = showTree;
    }
}
