/* ------------------------------------------------------------------------
 * Copyright 2012 Tim Vernum
 * ------------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ------------------------------------------------------------------------
 */
package org.adjective.syntactic.parser.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;

public class ModifierSet
{
    public static enum Modifier
    {
        PUBLIC(java.lang.reflect.Modifier.PUBLIC),
        PROTECTED(java.lang.reflect.Modifier.PROTECTED),
        PRIVATE(java.lang.reflect.Modifier.PRIVATE),
        ABSTRACT(java.lang.reflect.Modifier.ABSTRACT),
        STATIC(java.lang.reflect.Modifier.STATIC),
        FINAL(java.lang.reflect.Modifier.FINAL),
        SYNCHRONIZED(java.lang.reflect.Modifier.SYNCHRONIZED),
        NATIVE(java.lang.reflect.Modifier.NATIVE),
        TRANSIENT(java.lang.reflect.Modifier.TRANSIENT),
        VOLATILE(java.lang.reflect.Modifier.VOLATILE),
        STRICTFP(java.lang.reflect.Modifier.STRICT);

        private final int _mask;

        Modifier(final int javaBitMask)
        {
            _mask = javaBitMask;
        }

        public int getMask()
        {
            return _mask;
        }
    }

    private final EnumSet<Modifier> _modifiers;

    public ModifierSet()
    {
        _modifiers = EnumSet.noneOf(Modifier.class);
    }

    public ModifierSet(ModifierSet modifiers)
    {
        _modifiers = EnumSet.copyOf(modifiers._modifiers);
    }

    public ModifierSet(final Modifier... modifiers)
    {
        this(Arrays.asList(modifiers));
    }

    public ModifierSet(Collection<Modifier> modifiers)
    {
        _modifiers = EnumSet.copyOf(modifiers);
    }

    public ModifierSet(final int javaModifierBitmask)
    {
        this();
        for (Modifier modifier : Modifier.values())
        {
            if ((javaModifierBitmask & modifier.getMask()) != 0)
            {
                this.set(modifier);
            }
        }
    }

    public void set(Modifier modifier)
    {
        _modifiers.add(modifier);
    }

    public void clear(Modifier modifier)
    {
        _modifiers.remove(modifier);
    }

    public boolean is(Modifier modifier)
    {
        return _modifiers.contains(modifier);
    }

    public boolean isEmpty()
    {
        return _modifiers.isEmpty();
    }

    public String toString()
    {
        if (_modifiers.isEmpty())
        {
            return "";
        }
        StringBuilder buf = new StringBuilder();
        for (Modifier modifier : Modifier.values())
        {
            if (is(modifier))
            {
                buf.append(modifier.name().toLowerCase()).append(' ');
            }
        }
        buf.deleteCharAt(buf.length() - 1);
        return buf.toString();
    }
}
