/*
 * ------------------------------------------------------------------------
 *  Copyright 2012-2013 Tim Vernum
 *  ------------------------------------------------------------------------
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  ------------------------------------------------------------------------
 */
package org.adjective.syntactic.parser.util;

import org.adjective.syntactic.parser.ast.*;
import org.adjective.syntactic.parser.jj.Token;
import org.adjective.syntactic.parser.node.ExpressionSuffix;
import org.adjective.syntactic.parser.node.HasAnnotations;
import org.adjective.syntactic.parser.node.HasModifiers;
import org.adjective.syntactic.parser.node.HasTokens;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

public class PrintSourceVisitor implements JavaParserVisitor<Object, PrintStream>
{
    private long _indent = 0;

    public void print(final ASTCompilationUnit ast, final OutputStream output)
    {
        ast.jjtAccept(this, new PrintStream(output));
    }

    private CharSequence getIndentString()
    {
        final StringBuilder builder = new StringBuilder();
        return printIndent(builder);
    }

    private <T extends Appendable> T printIndent(T out)
    {
        for (int i = 0; i < _indent; i++)
        {
            try
            {
                out.append("  ");
            }
            catch (IOException e)
            {
                // swallow it...
            }
        }
        return out;
    }

    protected Object visitNode(Node node, PrintStream out)
    {
        return node.jjtAccept(this, out);
    }

    private boolean printSource(PrintStream out, Node node)
    {
        if (node == null)
        {
            return false;
        }
        visitNode(node, out);
        return true;
    }

    private <T extends HasAnnotations & HasTokens> Object printAnnotationSource(final PrintStream out, final T node)
    {
        printSpecial(out, node);
        return printAnnotationSource(out, node.getAnnotations());
    }

    private Object printAnnotationSource(final PrintStream out, final Iterable<ASTAnnotation> annotations)
    {
        for (ASTAnnotation annotation : annotations)
        {
            visitNode(annotation, out);
            out.println();
        }
        return null;
    }

    private <T extends HasAnnotations & HasModifiers & HasTokens> void printAnnotationsAndModifiers(PrintStream out, T node)
    {
        printAnnotationSource(out, node);
        ModifierSet modifiers = node.getModifiers();
        if (modifiers == null)
        {
            throw new IllegalArgumentException("Node: " + node + " has null modifiers");
        }
        if (!modifiers.isEmpty())
        {
            out.print(modifiers);
            out.print(' ');
        }
    }

    private Object printListSource(final PrintStream out, final Iterable<? extends Node> list, final CharSequence separator)
    {
        boolean first = true;
        for (Node node : list)
        {
            if (first)
            {
                first = false;
            }
            else
            {
                out.print(separator);
            }
            visitNode(node, out);
        }
        return null;
    }

    private void printSpecial(final PrintStream out, final HasTokens node)
    {
        final Token token = node.jjtGetFirstToken();
        if (token == null)
        {
            return;
        }
        for (Token special = token.specialToken; special != null; special = special.next)
        {
            out.println(special);
        }
    }

    @Override
    public Object visit(final SimpleNode node, final PrintStream out)
    {
        throw new UnsupportedOperationException("Attempt to visit node '"
                                                        + node.getClass().getSimpleName() + "' as SimpleNode");
    }

    @Override
    public Object visit(final ASTCompilationUnit node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTPackageDeclaration node, final PrintStream out)
    {
        node.getAnnotationsNode().jjtAccept(this, out);
        out.print("package ");
        node.getPackageName().jjtAccept(this, out);
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTAnnotations node, final PrintStream out)
    {
        return printAnnotationSource(out, node);
    }

    @Override
    public Object visit(final ASTImports node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTImportDeclaration node, final PrintStream out)
    {
        out.print("import ");
        node.childrenAccept(this, out);
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTStaticImport node, final PrintStream out)
    {
        out.print(" static ");
        return null;
    }

    @Override
    public Object visit(final ASTWildcardImport node, final PrintStream out)
    {
        out.print(".*");
        return null;
    }

    @Override
    public Object visit(final ASTTypes node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTModifiers node, final PrintStream out)
    {
        out.print(node.getModifiers());
        return null;
    }

    @Override
    public Object visit(final ASTTypeDeclaration node, final PrintStream out)
    {
        if (node.isEmpty())
        {
            out.println(";");
            return null;
        }

        printAnnotationsAndModifiers(out, node);
        return node.getDeclarationBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTClassOrInterfaceDeclaration node, final PrintStream out)
    {
        if (node.isInterface())
        {
            out.print("interface");
        }
        else
        {
            out.print("class");
        }
        out.print(' ');
        node.getName().jjtAccept(this, out);
        printSource(out, node.getTypeParameters());
        out.print(' ');
        if (printSource(out, node.getExtends()))
        {
            out.print(' ');
        }
        if (printSource(out, node.getImplements()))
        {
            out.print(' ');
        }
        return node.getBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTClassOrInterfaceKeyword node, final PrintStream out)
    {
        out.print(node.isInterface() ? "interface" : "class");
        return null;
    }

    @Override
    public Object visit(final ASTIdentifier node, final PrintStream out)
    {
        out.print(node.getIdentifier());
        return null;
    }

    @Override
    public Object visit(final ASTExtendsList node, final PrintStream out)
    {
        out.print("extends ");
        return printListSource(out, node.getTypes(), ", ");
    }

    @Override
    public Object visit(final ASTImplementsList node, final PrintStream out)
    {
        out.print("implements ");
        return printListSource(out, node.getTypes(), ", ");
    }

    @Override
    public Object visit(final ASTEnumDeclaration node, final PrintStream out)
    {
        out.print("enum ");
        node.getName().jjtAccept(this, out);
        out.print(' ');
        printSource(out, node.getImplementsNode());
        return node.getBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTEnumBody node, final PrintStream out)
    {
        out.print("{");
        node.getConstants().jjtAccept(this, out);
        out.println(";");
        return printListSource(out, node.getMembers(), "\n");
    }

    @Override
    public Object visit(final ASTEnumConstants node, final PrintStream out)
    {
        return printListSource(out, node.getConstants(), ", ");
    }

    @Override
    public Object visit(final ASTEnumConstant node, final PrintStream out)
    {
        printAnnotationsAndModifiers(out, node);
        node.getName().jjtAccept(this, out);
        printSource(out, node.getConstructorArguments());
        return printSource(out, node.getBody());
    }

    @Override
    public Object visit(final ASTTypeParameters node, final PrintStream out)
    {
        out.print("<");
        printListSource(out, node.getParameters(), ", ");
        out.print(">");
        return null;
    }

    @Override
    public Object visit(final ASTTypeParameter node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTTypeBound node, final PrintStream out)
    {
        out.print(" extends ");
        return printListSource(out, node.getBounds(), " & ");
    }

    @Override
    public Object visit(final ASTClassOrInterfaceBody node, final PrintStream out)
    {
        _indent++;
        out.println("{");
        if (!node.isEmpty())
        {
            printIndent(out);
            printListSource(out, node.getElements(), "");
        }
        _indent--;
        printIndent(out);
        out.println("}");
        return null;
    }

    @Override
    public Object visit(final ASTClassOrInterfaceBodyElement node, final PrintStream out)
    {
        if (node.isEmpty())
        {
            out.println(";");
        }
        else
        {
            node.childrenAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTMember node, final PrintStream out)
    {
        printAnnotationsAndModifiers(out, node);
        return node.getMember().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTFieldDeclaration node, final PrintStream out)
    {
        node.getDeclaredType().jjtAccept(this, out);
        out.print(' ');
        printListSource(out, node.getFields(), ", ");
        out.println(';');
        return null;
    }

    @Override
    public Object visit(final ASTVariableDeclarator node, final PrintStream out)
    {
        node.getIdentifier().jjtAccept(this, out);
        node.getArraySuffix().jjtAccept(this, out);
        if (node.hasInitializer())
        {
            out.print(" = ");
            node.getInitializer().jjtAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTArrayInitializer node, final PrintStream out)
    {
        out.print("{ ");
        printListSource(out, node.getElements(), ", ");
        out.print(" }");
        return null;
    }

    @Override
    public Object visit(final ASTMethodDeclaration node, final PrintStream out)
    {
        if (printSource(out, node.getTypeParameters()))
        {
            out.print(' ');
        }
        node.getDeclaredReturnType().jjtAccept(this, out);
        out.print(' ');
        node.getName().jjtAccept(this, out);
        node.getParameters().jjtAccept(this, out);
        node.getArraySuffix().jjtAccept(this, out);

        ASTNameList exceptions = node.getThrows();
        if (exceptions != null)
        {
            out.print(" throws ");
            visitNode(exceptions, out);
        }
        out.print(' ');
        return node.getBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTMethodBody node, final PrintStream out)
    {
        printSpecial(out, node);
        if (node.isEmpty())
        {
            out.println(";");
            return null;
        }
        if (node.isDefaultMethod())
        {
            out.print("default ");
        }
        return node.getBlock().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTArraySuffixList node, final PrintStream out)
    {
        for (int i = 0; i < node.getArrayDepth(); i++)
        {
            out.print("[]");
        }
        return null;
    }

    @Override
    public Object visit(final ASTArraySuffix node, final PrintStream out)
    {
        out.print("[]");
        return null;
    }

    @Override
    public Object visit(final ASTFormalParameters node, final PrintStream out)
    {
        out.print('(');
        printListSource(out, node.getParameters(), ", ");
        out.print(')');
        return null;
    }

    @Override
    public Object visit(final ASTFormalParameter node, final PrintStream out)
    {
        printAnnotationsAndModifiers(out, node);
        node.getTypeNode().jjtAccept(this, out);
        if (node.isVarArgs())
        {
            out.print("...");
        }
        out.print(' ');
        node.getName().jjtAccept(this, out);
        printSource(out, node.getArraySuffix());
        return null;
    }

    @Override
    public Object visit(final ASTVariableModifiers node, final PrintStream out)
    {
        printAnnotationsAndModifiers(out, node);
        return null;
    }

    @Override
    public Object visit(final ASTVarArgs node, final PrintStream out)
    {
        out.print("...");
        return null;
    }

    @Override
    public Object visit(final ASTConstructorDeclaration node, final PrintStream out)
    {
        if (printSource(out, node.getTypeParameters()))
        {
            out.print(' ');
        }
        node.getName().jjtAccept(this, out);
        node.getParameters().jjtAccept(this, out);
        if (node.hasExceptions())
        {
            out.print(" throws ");
            printSource(out, node.getExceptions());
        }
        return node.getBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTConstructorBody node, final PrintStream out)
    {
        out.println("{");
        node.childrenAccept(this, out);
        out.println("}");
        return null;
    }

    @Override
    public Object visit(final ASTExplicitConstructorInvocation node, final PrintStream out)
    {
        if (this.printSource(out, node.getContextExpression()))
        {
            out.print(".");
        }
        this.printSource(out, node.getTypeArguments());
        node.getTarget().jjtAccept(this, out);
        node.getArguments().jjtAccept(this, out);
        out.print(";");
        return null;
    }

    @Override
    public Object visit(final ASTInitializer node, final PrintStream out)
    {
        if (node.isStaticInitializer())
        {
            out.print("static ");
        }
        return node.getBlock().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTReferenceType node, final PrintStream out)
    {
        node.getComponentType().jjtAccept(this, out);
        for (int i = 0; i < node.getArrayDepth(); i++)
        {
            out.print("[]");
        }
        return null;
    }

    @Override
    public Object visit(final ASTClassOrInterfaceType node, final PrintStream out)
    {
        return printListSource(out, node.getComponents(), ".");
    }

    @Override
    public Object visit(final ASTParameterizedIdentifier node, final PrintStream out)
    {
        node.getIdentifier().jjtAccept(this, out);
        printSource(out, node.getTypeArguments());
        return null;
    }

    @Override
    public Object visit(final ASTTypeArguments node, final PrintStream out)
    {
        if (node.isEmpty())
        {
            out.print("<>");
        }
        else
        {
            out.print("<");
            printListSource(out, node.getArguments(), ", ");
            out.print(">");
        }
        return null;
    }

    @Override
    public Object visit(final ASTTypeArgument node, final PrintStream out)
    {
        if (node.isEmpty())
        {
            out.print("?");
        }
        else
        {
            if (node.isWildcard())
            {
                out.print("? ");
            }
            node.childrenAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTWildcardBounds node, final PrintStream out)
    {
        if (node.isExtendsBound())
        {
            out.print("extends ");
        }
        else
        {
            out.print("super ");
        }
        return node.getBound().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTPrimitiveType node, final PrintStream out)
    {
        out.print(node.getTypeName());
        return null;
    }

    @Override
    public Object visit(final ASTVoidType node, final PrintStream out)
    {
        out.print("void");
        return null;
    }

    @Override
    public Object visit(final ASTQualifiedIdentifier node, final PrintStream out)
    {
        return printListSource(out, node.getIdentifierComponents(), ".");
    }

    @Override
    public Object visit(final ASTNameList node, final PrintStream out)
    {
        return printListSource(out, node.getNames(), ", ");
    }

    @Override
    public Object visit(final ASTExpression node, final PrintStream out)
    {
        node.getBaseExpression().jjtAccept(this, out);
        if (node.isAssignment())
        {
            node.getOperator().jjtAccept(this, out);
            node.getAssignmentRHS().jjtAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTAssignmentOperator node, final PrintStream out)
    {
        String text = getOperatorText(node.getOperator());
        out.print(text);
        return null;
    }

    private String getOperatorText(final ASTAssignmentOperator.AssignmentOperator operator)
    {
        switch (operator)
        {
            case ASSIGN:
                return "=";
            case MULTIPLY_ASSIGN:
                return "*=";
            case DIVIDE_ASSIGN:
                return "/=";
            case MODULUS_ASSIGN:
                return "%=";
            case ADD_ASSIGN:
                return "+=";
            case SUBTRACT_ASSIGN:
                return "-=";
            case LEFT_SHIFT_ASSIGN:
                return "<<=";
            case RIGHT_SIGNED_SHIFT_ASSIGN:
                return ">>=";
            case RIGHT_UNSIGNED_SHIFT_ASSIGN:
                return ">>>=";
            case AND_ASSIGN:
                return "&=";
            case XOR_ASSIGN:
                return "^=";
            case OR_ASSIGN:
                return "|=";
        }
        throw new IllegalArgumentException("Unexpected assignment operator: " + operator);
    }

    @Override
    public Object visit(final ASTTernaryExpression node, final PrintStream out)
    {
        node.getCondition().jjtAccept(this, out);
        out.print("?");
        node.getTrueExpression().jjtAccept(this, out);
        out.print(":");
        node.getFalseExpression().jjtAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTLogicalOrExpression node, final PrintStream out)
    {
        return printListSource(out, node.getExpressions(), " || ");
    }

    @Override
    public Object visit(final ASTLogicalAndExpression node, final PrintStream out)
    {
        return printListSource(out, node.getExpressions(), " && ");
    }

    @Override
    public Object visit(final ASTInclusiveOrExpression node, final PrintStream out)
    {
        return printListSource(out, node.getExpressions(), " | ");
    }

    @Override
    public Object visit(final ASTExclusiveOrExpression node, final PrintStream out)
    {
        return printListSource(out, node.getExpressions(), " ^ ");
    }

    @Override
    public Object visit(final ASTAndExpression node, final PrintStream out)
    {
        return printListSource(out, node.getExpressions(), " & ");
    }

    @Override
    public Object visit(final ASTEqualityExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTBinaryOperator node, final PrintStream out)
    {
        out.print(getOperatorText(node.getOperator()));
        return null;
    }

    private String getOperatorText(final ASTBinaryOperator.BinaryOperator operator)
    {
        switch (operator)
        {
            case RIGHT_UNSIGNED_SHIFT:
                return ">>>";
            case RIGHT_SIGNED_SHIFT:
                return ">>";
            case LEFT_SHIFT:
                return "<<";
            case EQUAL:
                return "==";
            case NOT_EQUAL:
                return "!=";
            case LESS_THAN:
                return "<";
            case LESS_EQUAL:
                return "<=";
            case GREATER_THAN:
                return ">";
            case GREATER_EQUAL:
                return ">=";
            case PLUS:
                return "+";
            case MINUS:
                return "-";
            case MULTIPLY:
                return "*";
            case DIVIDE:
                return "/";
            case MODULUS:
                return "%";
        }
        throw new IllegalArgumentException("Unexpected binary operator: " + operator);
    }

    @Override
    public Object visit(final ASTInstanceOfExpression node, final PrintStream out)
    {
        node.getExpression().jjtAccept(this, out);
        out.print(" instanceof ");
        return node.getType().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTRelationalExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTShiftExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTAdditiveExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTMultiplicativeExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTUnaryExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTPrefixExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTUnaryOperator node, final PrintStream out)
    {
        String text = getOperatorText(node.getOperator());
        out.print(text);
        return null;
    }

    private String getOperatorText(final ASTUnaryOperator.UnaryOperator operator)
    {
        switch (operator)
        {
            case BINARY_NOT:
                return "~";
            case LOGICAL_NOT:
                return "!";
            case NEGATIVE:
                return "-";
            case POSITIVE:
                return "+";
            case DECREMENT:
                return "--";
            case INCREMENT:
                return "++";
        }
        throw new IllegalArgumentException("Unexpected operator " + operator);

    }

    @Override
    public Object visit(final ASTPostfixExpression node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTCastExpression node, final PrintStream out)
    {
        out.print("(");
        node.getType().jjtAccept(this, out);
        out.print(")");
        return node.getExpression().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTPrimaryExpression node, final PrintStream out)
    {
        node.getPrefix().jjtAccept(this, out);
        for (ExpressionSuffix suffix : node.getSuffixes())
        {
            visitNode(suffix, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTSuperExpression node, final PrintStream out)
    {
        out.print("super.");
        return node.getExpression().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTBracketExpression node, final PrintStream out)
    {
        out.print("(");
        node.getExpression().jjtAccept(this, out);
        out.print(")");
        return null;
    }

    @Override
    public Object visit(final ASTMemberSuffix node, final PrintStream out)
    {
        out.print(".");
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTIdentifierSuffix node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTMemberReference node, final PrintStream out)
    {
        out.print("::");
        printSource(out, node.getTypeArguments());
        if (node.isConstructorReference())
        {
            out.print("new");
        }
        else
        {
            node.getName().jjtAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTNew node, final PrintStream out)
    {
        out.print("new");
        return null;
    }

    @Override
    public Object visit(final ASTIntegerLiteral node, final PrintStream out)
    {
        out.print(node.getValue());
        return null;
    }

    @Override
    public Object visit(final ASTFloatingPointLiteral node, final PrintStream out)
    {
        out.print(node.getValue());
        return null;
    }

    @Override
    public Object visit(final ASTCharacterLiteral node, final PrintStream out)
    {
        out.print('\'');
        printCharacter(out, node.getCharacter());
        out.print('\'');
        return null;
    }

    private void printCharacter(final PrintStream out, final char ch)
    {
        if (ch >= 0x20 && ch <= 0x7e)
        {
            if (ch == '\'' || ch == '\\')
            {
                out.print('\\');
            }
            out.print(ch);
        }
        else if (ch == '\n')
        {
            out.print("\\n");

        }
        else if (ch == '\t')
        {
            out.print("\\t");

        }
        else if (ch <= 255)
        {
            out.print('\\');
            String str = "00" + Integer.toOctalString(ch);
            out.print(str.substring(str.length() - 3));
        }
        else
        {
            out.print("\\u");
            String str = "00" + Integer.toHexString(ch);
            out.print(str.substring(str.length() - 4));
        }
    }

    @Override
    public Object visit(final ASTStringLiteral node, final PrintStream out)
    {
        out.print(node.getImage());
        return null;
    }

    @Override
    public Object visit(final ASTThis node, final PrintStream out)
    {
        out.print("this");
        return null;
    }

    @Override
    public Object visit(final ASTSuper node, final PrintStream out)
    {
        out.print("super");
        return null;
    }

    @Override
    public Object visit(final ASTClassReference node, final PrintStream out)
    {
        node.getType().jjtAccept(this, out);
        out.print(".class");
        return null;
    }

    @Override
    public Object visit(final ASTArrayIndex node, final PrintStream out)
    {
        out.print("[");
        node.getExpression().jjtAccept(this, out);
        out.print("]");
        return null;
    }

    @Override
    public Object visit(final ASTBooleanLiteral node, final PrintStream out)
    {
        out.print(node.getValue());
        return null;
    }

    @Override
    public Object visit(final ASTNullLiteral node, final PrintStream out)
    {
        out.print("null");
        return null;
    }

    @Override
    public Object visit(final ASTArguments node, final PrintStream out)
    {
        out.print("(");
        if (!node.isEmpty())
        {
            node.getArguments().jjtAccept(this, out);
        }
        out.print(")");
        return null;
    }

    @Override
    public Object visit(final ASTArgumentList node, final PrintStream out)
    {
        printListSource(out, node.getArguments(), ", ");
        return null;
    }

    @Override
    public Object visit(final ASTAllocationExpression node, final PrintStream out)
    {
        out.print("new ");
        node.getAllocationType().jjtAccept(this, out);
        printSource(out, node.getTypeArguments());
        if (node.isArray())
        {
            node.getArrayNode().jjtAccept(this, out);
        }
        else
        {
            node.getConstructorArguments().jjtAccept(this, out);
            if (node.isAnonymousClass())
            {
                out.print(' ');
                node.getClassBody().jjtAccept(this, out);
            }
        }
        return null;
    }

    @Override
    public Object visit(final ASTArrayDimensions node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTDefinedArrayDimension node, final PrintStream out)
    {
        out.print("[");
        node.getDimension().jjtAccept(this, out);
        out.print("]");
        return null;
    }

    @Override
    public Object visit(final ASTUndefinedArrayDimension node, final PrintStream out)
    {
        out.print("[]");
        return null;
    }

    @Override
    public Object visit(final ASTArrayWithInit node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTExpressionStatement node, final PrintStream out)
    {
        node.getExpression().jjtAccept(this, out);
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTAssertStatement node, final PrintStream out)
    {
        out.print("assert ");
        node.getAssertion().jjtAccept(this, out);
        if (node.hasMessage())
        {
            out.print(" : ");
            node.getMessage().jjtAccept(this, out);
        }
        out.print(";");
        return null;
    }

    @Override
    public Object visit(final ASTLabeledStatement node, final PrintStream out)
    {
        node.getLabel().jjtAccept(this, out);
        out.print(": ");
        return node.getStatement().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTBlock node, final PrintStream out)
    {
        out.println("{");
        _indent++;
        if (!node.isEmpty())
        {
            printIndent(out);
            printListSource(out, node.statements(), getIndentString());
        }
        _indent--;
        printIndent(out);
        out.println("}");
        return null;
    }

    @Override
    public Object visit(final ASTVariableDeclarationStatement node, final PrintStream out)
    {
        node.getDeclaration().jjtAccept(this, out);
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTLocalVariableDeclaration node, final PrintStream out)
    {
        printAnnotationsAndModifiers(out, node);
        node.getTypeNode().jjtAccept(this, out);
        out.print(' ');
        printListSource(out, node.getVariables(), ", ");
        return null;
    }

    @Override
    public Object visit(final ASTEmptyStatement node, final PrintStream out)
    {
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTStatementExpression node, final PrintStream out)
    {
        node.getBaseExpression().jjtAccept(this, out);
        if (node.isPostfixExpression())
        {
            node.getPostfixOperator().jjtAccept(this, out);
        }
        else if (node.isAssignmentExpression())
        {
            out.print(' ');
            node.getAssignmentOperator().jjtAccept(this, out);
            out.print(' ');
            node.getAssignmentRHS().jjtAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTSwitchStatement node, final PrintStream out)
    {
        out.print("switch( ");
        node.getSwitchExpression().jjtAccept(this, out);
        out.println(") {");
        printListSource(out, node.getSwitchCases(), "\n");
        out.println("}");
        return null;
    }

    @Override
    public Object visit(final ASTSwitchCase node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTSwitchLabel node, final PrintStream out)
    {
        if (node.isDefault())
        {
            out.println("default:");
        }
        else
        {
            out.print("case ");
            node.getCaseExpression().jjtAccept(this, out);
            out.println(":");
        }
        return null;
    }

    @Override
    public Object visit(final ASTIfStatement node, final PrintStream out)
    {
        out.print("if(");
        node.getCondition().jjtAccept(this, out);
        out.print(") ");
        node.getIfBody().jjtAccept(this, out);
        if (node.hasElse())
        {
            out.print("else ");
            node.getElseBody().jjtAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTWhileStatement node, final PrintStream out)
    {
        out.print("while(");
        node.getCondition().jjtAccept(this, out);
        out.println(")");
        return node.getBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTDoStatement node, final PrintStream out)
    {
        out.print("do");
        node.getBody().jjtAccept(this, out);
        out.println("while(");
        node.getCondition().jjtAccept(this, out);
        out.println(");");
        return null;
    }

    @Override
    public Object visit(final ASTForStatement node, final PrintStream out)
    {
        out.print("for(");
        node.getCondition().jjtAccept(this, out);
        out.println(")");
        node.getStatement().jjtAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTForEach node, final PrintStream out)
    {
        printAnnotationsAndModifiers(out, node);
        node.getTypeNode().jjtAccept(this, out);
        out.print(' ');
        node.getIdentifier().jjtAccept(this, out);
        out.print(" : ");
        node.getIterable().jjtAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTTernaryForCondition node, final PrintStream out)
    {
        printSource(out, node.getInit());
        out.print(" ; ");
        printSource(out, node.getCondition());
        out.print(" ; ");
        printSource(out, node.getUpdate());
        return null;
    }

    @Override
    public Object visit(final ASTStatementExpressionList node, final PrintStream out)
    {
        return printListSource(out, node.getExpressions(), ", ");
    }

    @Override
    public Object visit(final ASTBreakStatement node, final PrintStream out)
    {
        out.print("break");
        if (!node.isEmpty())
        {
            out.print(' ');
            node.getLabel().jjtAccept(this, out);
        }
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTContinueStatement node, final PrintStream out)
    {
        out.print("continue");
        if (!node.isEmpty())
        {
            out.print(' ');
            node.getLabel().jjtAccept(this, out);
        }
        out.println(";");
        return null;

    }

    @Override
    public Object visit(final ASTReturnStatement node, final PrintStream out)
    {
        out.print("return ");
        node.childrenAccept(this, out);
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTThrowStatement node, final PrintStream out)
    {
        out.print("throw ");
        node.getExceptionExpression().jjtAccept(this, out);
        out.println(";");
        return null;
    }

    @Override
    public Object visit(final ASTSynchronizedStatement node, final PrintStream out)
    {
        out.print("synchronized(");
        node.getMonitorExpression().jjtAccept(this, out);
        out.print(")");
        return node.getBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTTryStatement node, final PrintStream out)
    {
        out.print("try ");
        printSource(out, node.getResourcesSpecification());
        node.getBody().jjtAccept(this, out);
        node.getCatchClauses().jjtAccept(this, out);
        if (node.hasFinallyBlock())
        {
            out.print(" finally ");
            node.getFinallyBlock().jjtAccept(this, out);
        }
        return null;
    }

    @Override
    public Object visit(final ASTResourcesSpecification node, final PrintStream out)
    {
        out.print("(");
        printListSource(out, node.getResources(), "; ");
        out.print(")");
        return null;
    }

    @Override
    public Object visit(final ASTResource node, final PrintStream out)
    {
        printAnnotationsAndModifiers(out, node);
        node.getDeclaredType().jjtAccept(this, out);
        out.print(' ');
        node.getName().jjtAccept(this, out);
        node.getArraySuffix().jjtAccept(this, out);
        out.print(" = ");
        return node.getAssignmentValue().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTCatchClauses node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTCatchClause node, final PrintStream out)
    {
        out.print("catch(");
        node.getParameter().jjtAccept(this, out);
        out.print(")");
        return node.getBody().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTCatchParameter node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTCatchType node, final PrintStream out)
    {
        return printListSource(out, node.getExceptionTypes(), " | ");
    }

    @Override
    public Object visit(final ASTNormalAnnotation node, final PrintStream out)
    {
        out.print('@');
        node.getName().jjtAccept(this, out);
        out.print('(');
        node.getValuePairs().jjtAccept(this, out);
        out.print(')');
        return null;
    }

    @Override
    public Object visit(final ASTMarkerAnnotation node, final PrintStream out)
    {
        out.print('@');
        return node.getName().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTSingleMemberAnnotation node, final PrintStream out)
    {
        out.print('@');
        node.getName().jjtAccept(this, out);
        out.print('(');
        node.getValue().jjtAccept(this, out);
        out.print(')');
        return null;
    }

    @Override
    public Object visit(final ASTMemberValuePairs node, final PrintStream out)
    {
        return printListSource(out, node.getPairs(), ", ");
    }

    @Override
    public Object visit(final ASTMemberValuePair node, final PrintStream out)
    {
        node.getName().jjtAccept(this, out);
        out.print("=");
        return node.getValue().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTMemberValue node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTMemberValueArrayInitializer node, final PrintStream out)
    {
        out.print("{");
        printListSource(out, node.getValues(), ", ");
        out.print(", }");
        return null;
    }

    @Override
    public Object visit(final ASTAnnotationTypeDeclaration node, final PrintStream out)
    {
        out.print("@interface");
        node.childrenAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTAnnotationTypeBody node, final PrintStream out)
    {
        out.println("{");
        node.childrenAccept(this, out);
        out.println("}");
        return null;
    }

    @Override
    public Object visit(final ASTAnnotationTypeMemberDeclaration node, final PrintStream out)
    {
        if (node.isEmpty())
        {
            out.println(";");
            return null;
        }
        printAnnotationsAndModifiers(out, node);
        return node.getMember().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTAnnotationMethodDeclaration node, final PrintStream out)
    {
        node.getType().jjtAccept(this, out);
        node.getName().jjtAccept(this, out);
        out.print("()");
        printSource(out, node.getDefaultValue());
        out.print(";");
        return null;
    }

    @Override
    public Object visit(final ASTDefaultValue node, final PrintStream out)
    {
        out.print("default ");
        return node.getValue().jjtAccept(this, out);
    }

    @Override
    public Object visit(final ASTLambdaExpression node, final PrintStream out)
    {
        node.getParameters().jjtAccept(this, out);
        out.print(" -> ");
        node.getBody().jjtAccept(this, out);
        return null;
    }

    @Override
    public Object visit(final ASTLambdaParameters node, final PrintStream out)
    {
        if (node.isSingleParameter())
        {
            node.getSingleParameterNode().jjtAccept(this, out);
        }
        else
        {
            out.print("(");
            if (!node.isEmpty())
            {
                node.getParameterListNode().jjtAccept(this, out);
            }
            out.print(")");
        }
        return null;
    }

    @Override
    public Object visit(final ASTLambdaParameterList node, final PrintStream out)
    {
        return printListSource(out, node.getParameters(), ", ");
    }

    @Override
    public Object visit(final ASTLambdaBody node, final PrintStream out)
    {
        node.childrenAccept(this, out);
        return null;
    }

}
